﻿using System;
using System.Collections.Generic;
using System.Text;
using OAI.Server.Catalog;
using OAI.Util;

namespace OAI.Server.Verb {
	public class ListRecords : ServerVerb {
		private static readonly List<string> validParamNames1 = new List<string> { "verb", "from", "until", "set", "metadataPrefix" };
		private static readonly List<string> validParamNames2 = new List<string> { "verb", "resumptionToken" };
		private static readonly List<string> requiredParamNames1 = new List<string> { "verb", "metadataPrefix" };
		private static readonly List<string> requiredParamNames2 = new List<string> { "verb", "resumptionToken" };

		public static string Construct(IOAIHandler oaiHandler)
		{
			var baseURL = oaiHandler.GetBaseURL();

			var query = oaiHandler.GetQuery();

			var oldResumptionToken = query["resumptionToken"];
			var metadataPrefix = query["metadataPrefix"];

			if (metadataPrefix != null && metadataPrefix.Length == 0)
				metadataPrefix = null;

			var sb = new StringBuilder();
			sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>");
			sb.Append("<OAI-PMH xmlns=\"http://www.openarchives.org/OAI/2.0/\"");
			sb.Append(" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
			sb.Append(" xsi:schemaLocation=\"http://www.openarchives.org/OAI/2.0/");
			sb.Append(" http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd\">");
			sb.Append("<responseDate>");

			var buildDate = DateTime.Now;
			sb.Append(CreateResponseDate(buildDate));
			sb.Append("</responseDate>");

			RecordsList records;

			try {
				List<string> requiredParamNames;
				List<string> validParamNames;

				if (oldResumptionToken == null) {
					validParamNames = validParamNames1;
					requiredParamNames = requiredParamNames1;

					sb.Append(GetRequestElement(query, validParamNames, baseURL));

					var from = query["from"];
					var until = query["until"];

					if ((!string.IsNullOrEmpty(from) && from.Length > 10) ||
						(!string.IsNullOrEmpty(until) && until.Length > 10) ||
						(from != null && until != null && until.Length != from.Length)) {
						throw new BadArgumentException();
					}

					var granularity = oaiHandler.GetGranularity();
					from = OAIUtil.GetFinestFrom(from, granularity);
					until = OAIUtil.GetFinestUntil(until, granularity);

					if (from != null && until != null && from.Length != until.Length) {
						throw new BadArgumentException();
					}

					var set = query["set"];
					if (set != null) {
						if (set.Length == 0) {
							set = null;
						}

						if (!oaiHandler.SupportsMetadataPrefix(metadataPrefix)) {
							throw new CannotDisseminateFormatException(metadataPrefix);
						}
					}

					if (!oaiHandler.SupportsMetadataPrefix(metadataPrefix)) {
						throw new CannotDisseminateFormatException(metadataPrefix);
					}

					records = oaiHandler.GetRecordsList(from, until, set, metadataPrefix, buildDate);
				} else {
					validParamNames = validParamNames2;
					requiredParamNames = requiredParamNames2;

					sb.Append(GetRequestElement(query, validParamNames, baseURL));

					if (HasBadArguments(query, requiredParamNames, validParamNames)) {
						throw new BadArgumentException();
					}

					records = oaiHandler.GetRecordsList(oldResumptionToken, buildDate);
				}

				if (records != null) {
					if (HasBadArguments(query, requiredParamNames, validParamNames)) {
						throw new BadArgumentException();
					}

					sb.Append("<ListRecords>");
					if (records.Records != null) {
						foreach (var record in records.Records) {
							sb.Append(record);
						}
					}

					var newResumptionMap = records.ResumptionMap;

					if (newResumptionMap != null) {
						var newResumptionToken = newResumptionMap.ResumptionToken;
						var completeListSize = newResumptionMap.CompleteListSize;
						var cursor = newResumptionMap.Cursor;

						sb.Append("<resumptionToken");

						if (completeListSize != null) {
							sb.Append(" completeListSize=\"");
							sb.Append(completeListSize.Value);
							sb.Append("\"");
						}

						if (cursor != null) {
							sb.Append(" cursor=\"");
							sb.Append(cursor.Value);
							sb.Append("\"");
						}

						sb.Append(">");
						sb.Append(newResumptionToken);
						sb.Append("</resumptionToken>");
					} else if (oldResumptionToken != null) {
						sb.Append("<resumptionToken />");
					}

					sb.Append("</ListRecords>");
				}
			} catch (BadArgumentException e) {
				sb.Append(e.Message);
			} catch (NoRecordsMatchException e) {
				sb.Append(e.Message);
			} catch (CannotDisseminateFormatException e) {
				sb.Append(e.Message);
			} catch (BadResumptionTokenException e) {
				sb.Append(e.Message);
			}

			sb.Append("</OAI-PMH>");

#if DEBUG
			//Debug.WriteLine(string.Format("ListRecords.Construct: returning: {0}", sb));
#endif
			return sb.ToString();
		}
	}
}
