﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Reflection;
using System.Text;
using System.Web;
using OAI.Server.Verb;
using ServiceCore;

namespace OAICat {
	public class OAIHandler : System.Web.UI.Page {
		private static readonly ServiceEngine _engine = ServiceEngine.Instance;
		private static bool preLoadError;

		protected void Page_PreLoad(object sender, EventArgs e) {
			preLoadError = false;
			if (ServiceEngine.PropertiesPath != null) return;

			try {
				ServiceEngine.PropertiesPath = Server.MapPath("configuration");
				_engine.LoadConfiguration(false);
			} catch (HttpException ex) {
				preLoadError = true;
				Response.StatusCode = 500;
				Response.ContentType = "text/plain; charset=UTF-8";
				Response.Write(ex);
			} catch (ServiceException ex) {
				preLoadError = true;
				Response.StatusCode = 500;
				Response.ContentType = "text/plain; charset=UTF-8";
				Response.Write(ex);
			}
		}

		protected void Page_Load(object sender, EventArgs e) {
			if (preLoadError) {
				preLoadError = false;
				return;
			}

//			lock (ServiceEngine.Instance)
			{
				var query = new NameValueCollection();

				foreach (var key in Request.QueryString.AllKeys) {
					switch (key) {
						case "institution":
							_engine.Institution = Request.QueryString[key];
							break;
						case "service":
							_engine.Service = Request.QueryString[key];
							break;
						default:
							query[key] = Request.QueryString[key];
							break;
					}
				}

				_engine.Query = query;

				var serverName = Request.ServerVariables["SERVER_NAME"];
				var serverPort = Request.ServerVariables["SERVER_PORT"];

				if (serverPort != "80")
					serverName = string.Format("{0}:{1}", serverName, serverPort);

				var appPath = Request.ApplicationPath;

				if (_engine.Institution != null && _engine.Service != null) {
					_engine.BaseURL = string.Format("http://{0}{1}/{2}/{3}/", serverName, appPath, _engine.Institution, _engine.Service);
				} else {
					_engine.BaseURL = string.Format("http://{0}/{1}", serverName, Request.RawUrl);
				}

				var serverVerbs = ServerVerb.Verbs;

				Request.ContentEncoding = new UTF8Encoding();

				var then = DateTime.Now;
#if DEBUG
				foreach (string headerName in Request.Headers.Keys) {
					Debug.WriteLine(string.Format("{0}: {1}", headerName, Request.Headers[headerName]));
				}
#endif

				GZipEncodePage(Request, Response);

				try {
					var result = GetResult(Request, serverVerbs);
					var contentType = _engine.ContentType;

					if (string.IsNullOrEmpty(contentType))
						contentType = "text/xml; charset=UTF-8";

					Response.ContentType = contentType;
					Response.Write(result);
				} catch (ServiceException ex) {
#if DEBUG
					Debug.Write(ex.Message);
					Debug.Write(ex.StackTrace);
#endif
					Response.StatusCode = 500;
					Response.ContentType = "text/plain; charset=UTF-8";
					Response.Write(ex);

					if (ex.InnerException != null) {
#if DEBUG
						Debug.Write(ex.InnerException.Message);
						Debug.Write(ex.InnerException.StackTrace);
#endif
						Response.Write(ex.InnerException);
						Response.Write(ex.InnerException.Message);
						Response.Write(ex.InnerException.StackTrace);
					}
				} catch (Exception ex) {
#if DEBUG
					Debug.Write(ex.Message);
					Debug.Write(ex.StackTrace);
#endif
					Response.StatusCode = 500;
					Response.ContentType = "text/plain; charset=UTF-8";
					Response.Write(ex.Message);
					Response.Write(ex.StackTrace);

					if (ex.InnerException != null) {
#if DEBUG
						Debug.Write(ex.InnerException.Message);
						Debug.Write(ex.InnerException.StackTrace);
#endif
						Response.Write(ex.InnerException);
						Response.Write(ex.InnerException.Message);
						Response.Write(ex.InnerException.StackTrace);
					}
				}

				var now = DateTime.Now;
				System.Diagnostics.Trace.WriteLine(string.Format("{0}: {1}ms", Request.RawUrl, (now - then).TotalMilliseconds));
			}
		}

		private static bool IsGZipSupported(HttpRequest request) {
			var AcceptEncoding = request.Headers["Accept-Encoding"];
			return !string.IsNullOrEmpty(AcceptEncoding) && (AcceptEncoding.Contains("gzip") || AcceptEncoding.Contains("deflate"));
		}

		private static void GZipEncodePage(HttpRequest request, HttpResponse response) {
			if (!IsGZipSupported(request)) return;

			var AcceptEncoding = request.Headers["Accept-Encoding"];
			if (AcceptEncoding.Contains("gzip")) {
				response.Filter = new System.IO.Compression.GZipStream(response.Filter, System.IO.Compression.CompressionMode.Compress);
				response.AppendHeader("Content-Encoding", "gzip");
			} else {
				response.Filter = new System.IO.Compression.DeflateStream(response.Filter, System.IO.Compression.CompressionMode.Compress);
				response.AppendHeader("Content-Encoding", "deflate");
			}
		}

		private static string GetResult(HttpRequest request, Dictionary<string, Type> verbs) {
			var verb = request.Params["verb"];

#if DEBUG
			Debug.WriteLine(string.Format("OAIHandler.GetResult: verb=>{0}<", verb));
#endif

			Type verbClass;
			if (verb != null && verbs.ContainsKey(verb)) {
				verbClass = verbs[verb];
			} else {
#if DEBUG
				Debug.WriteLine("Verb not found among: ");
				foreach (var v in verbs.Keys) {
					Debug.WriteLine(v);
				}
#endif
				verbClass = typeof(BadVerb);
			}

			var result = (string)verbClass.InvokeMember("Construct", BindingFlags.InvokeMethod | BindingFlags.Static | BindingFlags.Public, null, null, new object[] { _engine });

#if DEBUG
			//			Debug.WriteLine(result);
#endif

			return result;
		}
	}
}
